/*
 *  stats.c - output statistic files (refs, errs, html)
 *  Copyright (c)2002 - 2004 Andrej  OM1CW
 *  Copyright (c)2005 Mike Karas OM4AA <mike.om4aa@gmail.com>
 *  Copyright (c)2009 Ladislav Vaiz OK1ZIA <ok1zia@nagano.cz>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include "osdep.h"
#include "debug.h"
#include "convert.h"
#include "compare.h"
#include "main.h"
#include "qrb.h"
#include "load.h"
#include "check.h"
#include "call.h"
#include "diplom.h"

void count_forcedwwl_refs(uint8_t band)
{
	struct CALL_INFO *c;
	int i;
	
	for (i = 0; i < forcedwwl_num; i++)
		if ((c = add_call(forced[i].call, forced[i].wwl, band)) && c->refs < 20000)
			c->refs += 20000;
}

void make_refs(uint8_t band)
{
	DEBUG_FUNC("make_refs");

	struct CALL_INFO *c = first_call[band];
	struct QSO_INFO *q;

	printf(".");

	while (c) {
		c->refs = 0;
		q = c->first_qso;		// delete all !!!
		while (q) {
			// q->ok_call->refs=0;
			q->r_call->refs = 0;	// !!!! CHANGED 8.11.02
			q = q->next;
		}
		c = c->next;
	}

	c = first_call[band];

	while (c) {
		q = c->first_qso;
		if (q)
			c->refs += 10000;
		while (q) {
			// q->ok_call->refs++;
			q->r_call->refs++;	// !!!! CHANGED 8.11.02
			q = q->next;
		}
		c = c->next;
	}
	count_forcedwwl_refs(band);
}

void make_band_refs(uint8_t band)
{
	DEBUG_FUNC("make_band_refs");

	struct CALL_INFO *c = first_call[band], **c2;
	struct QSO_INFO *q, **q2;
	FILE *out;
	char filename[BUFF_SIZE + 1], text[BUFF_SIZE + 1], loc[7];
	float bnd = band2freq[band];
	int num = 0, i;
	char cc;

	printf(".");

	while (c) {					// clear refs
		c->refs = 0;
		q = c->first_qso;
		while (q) {
			q->r_call->refs = 0;
			q = q->next;
		}
		c = c->next;
	}

	c = first_call[band];		// make refs and num

	while (c) {
		q = c->first_qso;
		while (q) {
			if (!q->r_call->refs)
				num++;
			q->r_call->refs++;
			q = q->next;
		}
		c = c->next;
	}

	c = first_call[band];		// clear refs

	while (c) {
		q = c->first_qso;
		while (q) {
			q->r_call->refs = 0;
			q = q->next;
		}
		c = c->next;
	}


	q2 = (struct QSO_INFO **) malloc(num * sizeof(q));	// make refs again 
	// and
	// fill sort buffer

	i = 0;
	c = first_call[band];
	while (c) {
		q = c->first_qso;
		while (q) {
			if (!q->r_call->refs)
				q2[i++] = q;
			q->r_call->refs++;
			q = q->next;
		}
		c = c->next;
	}

	qsort(q2, num, sizeof(q), compare_fcn2);

	if (band >= BAND_GHZ && band < BAND_GHZ_ZERO)
		bnd *= 1000;

	if (band >= BAND_GHZ_ZERO)
		sprintf(filename, "%s%slx_G%i.ref", cfg_epath, DIR_DELIMITER,
				(int) bnd);
	else
		sprintf(filename, "%s%slx_%04i.ref", cfg_epath, DIR_DELIMITER,
				(int) bnd);
	out = fopen(filename, "wt");
	if (!out)
		return;

	if (band >= BAND_GHZ) {
		if (band >= BAND_GHZ_ZERO)
			sprintf(text, "%2.0f GHz", band2freq[band]);
		else
			sprintf(text, "%2.1f GHz", band2freq[band]);
	} else
		sprintf(text, "%2.0f MHz", band2freq[band]);

	text[8] = 0;

	fprintf(out, "Referencies on %s band\n\n", text);
	fprintf(out, "Call            Loc.   Log Refs  Status\n");


	for (i = 0; i < num; i++) {
		num2loc(q2[i]->r_call->loc, loc);
		if (q2[i]->r_call->first_qso)
			cc = 'X';
		else
			cc = ' ';
		fprintf(out, "%-16s%6s %c%6i - ", q2[i]->r_call->call, loc, cc,
				q2[i]->r_call->refs);
		if (q2[i]->r_call != q2[i]->ok_call) {
			num2loc(q2[i]->ok_call->loc, loc);
			if (q2[i]->ok_call->first_qso)
				cc = 'X';
			else
				cc = ' ';
			fprintf(out, "REPAIR: %-16s%6s %c%6i\n", q2[i]->ok_call->call,
					loc, cc, q2[i]->ok_call->refs);
		} else {
			fprintf(out, "OK\n");
		}
	}

	free(q2);

	c = first_call[band];		// clear refs

	while (c) {
		c->refs = 0;
		c->temp = 0;
		c = c->next;
	}

	num = 0;
	c = first_call[band];		// refs=num of QSO with log
	// temp=num of err QSO with log
	while (c) {
		q = c->first_qso;
		if (q && c->categ)
			num++;
		while (q) {
			if (q->ok_call->first_qso) {
				q->ok_call->refs++;
				if (!good_qso(q))
					q->ok_call->temp++;
			}
			q = q->next;
		}
		c = c->next;
	}

	c2 = (struct CALL_INFO **) malloc(num * sizeof(c));
	i = 0;

	c = first_call[band];

	while (c) {
		if (c->first_qso && c->categ)
			c2[i++] = c;
		c = c->next;
	}

	qsort(c2, num, sizeof(c), compare_fcn3);

	fprintf(out, "\nValid contacts of counterpart stations:\n\n");
	fprintf(out, "Call             QSOs  QSOs  Logged  Errors\n");
	fprintf(out, "                 all   log.  QSOs %%  %%\n");

	for (i = 0; i < num; i++) {
		fprintf(out, "%-16s %4i %4i   %5.1f   %5.1f\n", c2[i]->call,
				c2[i]->qsos_all, c2[i]->refs,
				c2[i]->qsos_all ? ((float) c2[i]->refs * 100 /
									c2[i]->qsos_all) : 0,
				c2[i]->refs ? ((float) c2[i]->temp * 100 /
									c2[i]->refs) : 0);
	}

	free(c2);

	fclose(out);
}

void make_err_files(uint8_t band, char *band_str)
{
	DEBUG_FUNC("make_err_files");

	struct CALL_INFO *c = first_call[band];
	struct QSO_INFO *q, *odx;
	char loc[7];
	uint32_t dis, score;
	/*uint32_t*/ int lost; /* was int */
	FILE *out;
	char filename[BUFF_SIZE + 1], *tmp, text[BUFF_SIZE + 1];
	uint32_t qsos;
	uint8_t err_rnz, err_nil, err_dupe10, err_wnc, err_irsts;

	// printf("+ Generating ERR files: ");

	printf(".");

	while (c) {
		q = c->first_qso;
		if (q && c->categ) {
			score = 0;
			lost = 0;
			qsos = 0;
			odx = NULL;

			sprintf(filename, "%s%s%s", cfg_epath, DIR_DELIMITER,
					c->filepath);
			make_path(filename);
			strcat(filename, c->filename);
			tmp = strrchr(filename, '.');
			if (tmp)
				*tmp = 0;
			sprintf(c->filename_err, "%s.err", filename);
			out = fopen(c->filename_err, "wt");

			if (!out) {
				fprintf(err, "! Error create error file: %s\n",
						c->filename_err);
			} else {

				num2loc(c->loc, loc);
				fprintf(out, "\n Error log: %s\n", ctest_name);
				fprintf(out, " Station:   %s\n", c->call);
				fprintf(out, " Locator:   %s\n", loc);
				fprintf(out, " Band:      %s\n", band_str);
				fprintf(out, " Category:  %s\n\n", categs[c->categ].name);
				fprintf(out,
						"                | Sent     | Received         | Correct data                    | Other | Lost\n");
				fprintf(out,
						" Call           | RST NR   | RST  NR   LOC    | Call           RST  NR   LOC    | error |points\n");
				fprintf(out,
						"-----------------------------------------------------------------------------------------------\n");

				err_rnz = 0;
				err_nil = 0;
				err_dupe10 = 0;
				err_wnc = 0;

				while (q) {
					dis = QRB(c->loc, q->r_call->loc);
					/*
					 * if (!q->dupe && dis!=q->claimed_QRB)
					 * dis=QRB(c->loc, q->r_call->loc); 
					 */
					q->QRB = 0;
					if (q->dupe);
					else if (good_qso(q)) {
						score += dis;
						q->QRB = dis;
						qsos++;
						if (!odx || odx->QRB < dis)
							odx = q;
					} else {
						num2loc(q->r_call->loc, loc);
						sprintf(text,
								" %-14s | %-3s %4d | %-3s %4d %7s | ",
								q->r_call->call, q->s_rst1, q->s_nr,
								q->r_rst1, q->r_nr, loc);
						if (!strcmp(q->ok_call->call, q->r_call->call))
							strcat(text, "               ");
						else
							sprintf(strchr(text, 0), "%-14s ",
									q->ok_call->call);

                        if (strcmp(q->ok_rst1, q->r_rst1)==0)
							strcat(text, "    ");
						else
							sprintf(strchr(text, 0), "%3s ", q->ok_rst1);

						if (q->ok_nr == q->r_nr)
							strcat(text, "     ");
						else
							sprintf(strchr(text, 0), "%4d ", q->ok_nr);
						num2loc(q->ok_call->loc, loc);
						if (q->ok_call->loc == q->r_call->loc)
							strcat(text, "        |");
						else
							sprintf(strchr(text, 0), "%7s |", loc);
						// ///////other error
						if (q->pass == 100) {
							strcat(text, " NIL   ");
							err_nil = 1;
						} else if (!q->ok_nr) {
							strcat(text, " RNZ   ");
							err_rnz = 1;
						} else if (!strcmp(q->r_call->call, q->ok_call->call)
							   && !q->dupe && q->claimed_QRB && q->pass == 101) {
							score -= dis * 9;	// if has correct call -> disqualification 10x
							dis *= 10;
							strcat(text, "DUPEx10");
							err_dupe10 = 1;
						} else if (q->pass == 101 || q->dupe) {
							if (!q->claimed_QRB)
								dis = 0; // no points lost (claimed 0)
							strcat(text, " DUPE  ");
							//sprintf(text+strlen(text), " DUPE q->pass=%d, q->dupe=%d ", q->pass, q->dupe);
						} else if (q->wwl_outside) {
							strcat(text, " WNC   ");
							err_wnc = 1;
						} else if (q->invalid_rsts) {
                            strcat(text, " IRSTS ");
                            err_irsts = 1;
                        } else
							strcat(text, "       ");


						sprintf(strchr(text, 0), "| -%4d\n", dis);
						fprintf(out, text);
						lost += dis;
					}
					q = q->next;
				}

				fprintf(out,
						"-----------------------------------------------------------------------------------------------\n");
				fprintf(out,
						"                                                                     Sum of lost points: -%5d\n",
						lost);
				if (err_rnz || err_nil || err_dupe10 || err_wnc) {
					fprintf(out, "\n Description of other errors:\n");
					if (err_rnz)
						fprintf(out, " RNZ = Received number is zero\n");
					if (err_nil)
						fprintf(out, " NIL = Not in the log\n");
					if (err_dupe10)
						fprintf(out,
								" DUPEx10 = Computed duplicate contact with good callsign. 10 x penalty\n");
					if (err_wnc)
						fprintf(out, " WNC = Received big WWL square is outside of the country\n");
                    if (err_irsts)
                        fprintf(out, " IRSTS = Invalid SENT RST\n");

				}
				fprintf(out, "\n Claimed score:   %6d", c->score_claimed);
				if (c->score_claimed == score + lost)
					fprintf(out, "\n");
				else
					fprintf(out,
							"  Incorrect claimed score and log summary (good=%d)\n",
							score + lost);
				fprintf(out, " Corrected score: %6d  Errors: %2.1f%%\n", score,
					(lost + score) ? ((float) (lost * 100)) / (lost + score) : 0);
                
                fprintf(out, "\n Checked by %s v%s\n", PACKAGE_NAME, PACKAGE_VERSION);
				/*
				 * if (c->score_claimed!=score+lost) { fprintf(out,
				 * "\nIncorrect computed QRB in QSOs:\n\nCall Loc Claimed
				 * QRB Correct QRB\n"); q=c->first_qso; while (q) {
				 * dis=QRB(c->loc, q->r_call->loc); if (!q->dupe &&
				 * dis!=q->claimed_QRB) { num2loc(q->r_call->loc, loc);
				 * fprintf(out, "%-14s %6s %4i %4i\n", q->r_call->call,
				 * loc, q->claimed_QRB, dis); } q=q->next; } } 
				 */
				c->score_lost = lost;
				c->score = score;
				c->odx = odx;
				c->qsos = qsos;
				/*
				 * fprintf(out, "\n For Your Information:\n");
				 * fprintf(out, " Call In other logs\n"); q=c->first_qso;
				 * while (q) { if (q->ok_call->refs<3) { fprintf(out, "
				 * %-20s %d ", q->r_call->call, q->r_call->refs-1);
				 * c2=q->ok_call->first_call_same_loc; while (c2) { if
				 * (c2!=q->ok_call && c2->refs) fprintf(out, "%s[%d] ",
				 * c2->call, c2->refs); c2=c2->next_call_same_loc; }
				 * fprintf(out, "\n"); }
				 * 
				 * q=q->next; } 
				 */
				fclose(out);
			}
			// printf(".");
		}
		c = c->next;
	}

	// printf("OK\n");
}

void make_score_list(uint8_t band, char *band_str, uint8_t categ, FILE * res,
	FILE * htm, FILE *fd, float maxpwr, int ignunkpwr)
{
	DEBUG_FUNC("make_score_list");

	char s[10], asl[10], pwr[50];
	struct CALL_INFO *c, **c2;
	int num = 0, i = 0;
	char loc[7], *callspace;
	char odx_call[SIZE_CALL + 1];
	int odx_QRB, band_freq;
	float average, errors;

	// printf("+ Generating %s files: ", filename);


	for (c = first_call[band]; c; c = c->next){
        if (c->pwr > maxpwr) continue;
        if (!ignunkpwr && c->pwr == 0.0) continue;

		if (c->categ == categ)
			num++;
	}

	if (!num)
		return;

	printf("+ Generating result of %s, %s: ", band_str,
		   categs[categ].name);

	if (band >= BAND_GHZ)
		band_freq = (int) (band2freq[band] * 1000);
	else
		band_freq = (int) band2freq[band];
	fprintf(fd, "\t<category band=\"%u\" operator=\"%s\">\n", band_freq, categs[categ].name);
	fprintf(htm,
			"<tt><font face=\"Arial CE\" size=5>Category: <i>%s, %s</i></font></tt><br><br>\n",
			band_str, categs[categ].name);
	// fprintf(htm, "<pre>\n");
	fprintf(htm, "<table BORDER BGCOLOR=\"#ffffff\">\n");

	fprintf(res, "\nCategory: %s, %s\n\n", band_str, categs[categ].name);

	c2 = (struct CALL_INFO **) malloc(num * sizeof(c));

	for (c = first_call[band]; c; c = c->next){
        if (c->pwr > maxpwr) continue;
        if (!ignunkpwr && c->pwr == 0.0) continue;

		if (c->categ == categ) {
			c2[i] = c;
			i++;
		}
	}

	qsort(c2, num, sizeof(c), compare_fcn);

	fprintf(res,
			"Pl. Call        score  QSO aver. LOC     ASL ODX         QRB   er. Power antenna\n");

	fprintf(htm, "<tr>\n");
	fprintf(htm, "<td><b><font face=\"Arial CE\">Pl.</font></b></td>\n");
	fprintf(htm, "<td><b><font face=\"Arial CE\">Call</font></b></td>\n");
	fprintf(htm, "<td><b><font face=\"Arial CE\">Score</font></b></td>\n");
	fprintf(htm, "<td><b><font face=\"Arial CE\">QSO</font></b></td>\n");
	fprintf(htm, "<td><b><font face=\"Arial CE\">Aver.</font></b></td>\n");
	fprintf(htm,
			"<td><b><font face=\"Arial CE\">Locator</font></b></td>\n");
	fprintf(htm, "<td><b><font face=\"Arial CE\">ASL</font></b></td>\n");
	fprintf(htm, "<td><b><font face=\"Arial CE\">ODX</font></b></td>\n");
	fprintf(htm, "<td><b><font face=\"Arial CE\">QRB</font></b></td>\n");
	fprintf(htm,
			"<td><b><font face=\"Arial CE\">Err. %%</font></b></td>\n");
	fprintf(htm, "<td><b><font face=\"Arial CE\">Power</font></b></td>\n");
	fprintf(htm,
			"<td><b><font face=\"Arial CE\">Antenna</font></b></td>\n");
	fprintf(htm, "</tr>\n");


	// fprintf(htm, "<b>Pl. Call score QSO aver. LOC ASL ODX QRB er. P[W]
	// antenna</b>\n");
	for (i = 0; i < num; i++) {
		num2loc(c2[i]->loc, loc);
		sprintf(s, "%i.", i + 1);

		if (strlen(s) < 4)
			strcat(s, space + strlen(space) + strlen(s) - 4);
		if (strlen(c2[i]->call) < 10)
			callspace = space + strlen(space) + strlen(c2[i]->call) - 10;
		else
			callspace = strchr(space, 0);

		if (c2[i]->asl)
			sprintf(asl, "%4i", c2[i]->asl);
		else
			strcpy(asl, "   ?");

		if (c2[i]->pwr)
			float_print(c2[i]->pwr, pwr);
        else
			strcpy(pwr, "   ?W");

		if (!*c2[i]->ant)
			strcpy(c2[i]->ant, "?");

		*odx_call = '\0';
		odx_QRB = 0;

		if (!c2[i]->odx) {
			strcpy(odx_call, "NONE");
			odx_QRB = 0;
		} else {
			strcpy(odx_call, c2[i]->odx->r_call->call);
			odx_QRB = c2[i]->odx->QRB;
		}
		
		average = (float) (c2[i]->qsos_all ? ((float) c2[i]->score) / c2[i]->qsos_all : 0);
		errors = (c2[i]->score_lost + c2[i]->score) ?
			(((float) (c2[i]->score_lost) * 100) / (c2[i]->score_lost + c2[i]->score)) : 0;

		fprintf(res,
				"%s%-10s %6i %4i %5.1f %s %s %-10s %4i %5.1f %s %-20s\n",
				s, c2[i]->call, c2[i]->score, c2[i]->qsos_all,
				(float) (c2[i]->qsos_all ? ((float) c2[i]->score) /
						 c2[i]->qsos_all : 0), loc, asl, odx_call, odx_QRB,
				(c2[i]->score_lost +
				 c2[i]->score) ? ((float) (c2[i]->score_lost) * 100) /
				(c2[i]->score_lost + c2[i]->score) : 0, pwr, c2[i]->ant);

		fprintf(htm, "<tr>\n");
		fprintf(htm,
				"<td align=right><b><font face=\"Arial CE\">%i.</font></b></td>\n",
				i + 1);
		fprintf(htm,
				"<td><b><font face=\"Arial CE\"><a href=\"%s\" target=\"_blank\">%s</a></font></b></td>\n",
				c2[i]->filename_err + strlen(cfg_epath) + 1, c2[i]->call);
		fprintf(htm,
				"<td align=right><b><font face=\"Arial CE\">%i</font></b></td>\n",
				c2[i]->score);
		fprintf(htm,
				"<td align=right><b><font face=\"Arial CE\">%i</font></b></td>\n",
				c2[i]->qsos_all);
		fprintf(htm,
				"<td align=right><b><font face=\"Arial CE\">%5.1f</font></b></td>\n",
				average);
		fprintf(htm,
				"<td><b><font face=\"Arial CE\">%s</font></b></td>\n", loc);
		fprintf(htm,
				"<td align=right><b><font face=\"Arial CE\">%s</font></b></td>\n",
				asl);
		fprintf(htm,
				"<td><b><font face=\"Arial CE\">%s</font></b></td>\n",
				odx_call);
		fprintf(htm,
				"<td align=right><b><font face=\"Arial CE\">%i</font></b></td>\n",
				odx_QRB);
		fprintf(htm,
				"<td align=right><b><font face=\"Arial CE\">%5.1f</font></b></td>\n",
				errors);
		fprintf(htm,
				"<td align=right><b><font face=\"Arial CE\">%s</font></b></td>\n",
				pwr);
		fprintf(htm,
				"<td><b><font face=\"Arial CE\">%s</font></b></td>\n",
				c2[i]->ant);
		fprintf(htm, "</tr>\n");

		/*
		 * fprintf(htm, "%s<a href=\"%s\">%s</a>%s %6i %4i %5.1f %s %s
		 * %-10s %4i %5.1f %s %-20s\n", s,
		 * c2[i]->filename_err+strlen(cfg_epath)+1, c2[i]->call,
		 * callspace, c2[i]->score, c2[i]->qsos,
		 * (float)(c2[i]->qsos?((float)c2[i]->score)/c2[i]->qsos:0), loc,
		 * asl, odx_call, odx_QRB,
		 * (c2[i]->score_lost+c2[i]->score)?((float)(c2[i]->score_lost)*100)/(c2[i]->score_lost+c2[i]->score):0, 
		 * pwr, c2[i]->ant); 
		 */
		 
		 fprintf(fd, "\t\t<station place=\"%u\" call=\"%s\" score=\"%u\" qsos=\"%u\" average=\"%.2f\" wwl=\"%s\" asl=\"%u\" odx=\"%s, %u\" errors=\"%.2f\" power=\"%.6f\" antenna=\"%s\" />\n",
		 	i + 1, c2[i]->call, c2[i]->score, c2[i]->qsos_all, average, loc, c2[i]->asl, odx_call, odx_QRB,
		 	errors, c2[i]->pwr, c2[i]->ant);

		 if (i+1 <= diplomplace)
			 make_diplom(categ, c2[i]->call, i+1, band_str);
	}

	// fclose(out);

	free(c2);
	
	fprintf(fd, "\t</category>\n");

	fprintf(htm, "</table><hr><br>");
	// fprintf(htm, "</pre>\n");
	fprintf(res, "\n");

	printf("%i stations\n", num);
}
